/* FILE: functions.cc                   -*-Mode: c++-*-
 *
 * Non-class C++ functions defined by the Nb extension.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 1998/05/31 21:42:02 $
 * Last modified by: $Author: donahue $
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "functions.h"

/* End includes */     

//////////////////////////////////////////////////////////////////////////
// Greatest common divisor via Euclid's algorithm
int Nb_Gcd(int m,int n)
{
  int temp;
  while((temp=(m%n))>0) { m=n; n=temp; }
  return n;
}

//////////////////////////////////////////////////////////////////////////
// Version of atof that allows 'D' or 'd' to be used to denote exponent
double Nb_Atof(const char *nptr)
{
  const char *cptr;
  if((cptr=strchr(nptr,'d'))==NULL &&
      (cptr=strchr(nptr,'D'))==NULL  )
    return atof(nptr);  // Usual, C-compatible string
  /// The above checks seem to add about 10% overhead for short
  /// strings on a 486/66dx2, and it is faster there to check for
  /// each of 'd' and 'D' separately than to call strcspn(nptr,"dD");
  /// The base atof time on this machine is about 85 microseconds
  /// to start with (compared to about 6 microseconds for atoi).

  // Otherwise, assume we are working with a FORTRAN-style
  // "double precision" string.
  //   On the aforementioned 486, with gcc 2.7.2, this code costs
  // on average about an additional 10% over the above, if statbuf
  // is used.  Calls to new[] and delete[] add another 20%.  NOTE:
  // the strncpy function is really slow on this machine (it takes
  // as long as the atof call itself), so instead of using strncpy
  // to ensure we don't write past the end of statbuf, we take a
  // more flexible approach: see how big the string is, and if it
  // is too big (which is unlikely, but possible), then allocate
  // temporary storage just large enough to hold the string.  The
  // latter is a performance hit, so if it happens frequently one
  // should enlarge the size of statbuf.
  int exp_off=cptr-nptr;
  static char statbuf[256]; // I don't expect many numeric strings this long
  char *buf=statbuf;
  size_t num_len=strlen(nptr)+1;
  if(num_len>sizeof(statbuf)) buf=new char[num_len];

  strcpy(buf,nptr);
  buf[exp_off]='e';         // Replace 'd' or 'D' with 'e'
  double result=atof(buf);

  if(buf!=statbuf) delete[] buf;
  return result;
}

//////////////////////////////////////////////////////////////////////////
// Portable case insensitive string comparison, modeled after
// strcasecmp().
int Nb_StrCaseCmp(const char *s1, const char *s2)
{
  const char *cptr1=s1,*cptr2=s2;
  int c1,c2;
  do {
    c1=tolower(*(cptr1++));
    c2=tolower(*(cptr2++));
  } while(c1==c2 && c1!='\0');
  if(c1<c2) return -1;
  if(c1>c2) return  1;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// C interface to the Tcl 'file nativename' command.
void Nb_TclNativeFilename(Tcl_Interp* interp,Nb_DString &filename)
{
  // Similar to the Tcl command 'file nativename', this routine provides
  // an interface to the Tcl C function Tcl_TranslateFileName(), using
  // NB_DString's instead of Tcl_DString's.  It converts "filename" from
  // Tcl standard format (with /'s) to a platform specific form.  If the
  // call is successful, then filename is overwritten with the returned
  // result. Otherwise filename is left unchanged.

  if(interp==NULL) {
    interp=Oc_GlobalInterpreter();
    if(interp==NULL) return;
  }

  /* Tcl_ResetResult(interp);  // Presumably this isn't needed. */

  Oc_AutoBuf ab(filename);
  Tcl_DString buf;
  Tcl_DStringInit(&buf); // Safety init
  char *newname=Tcl_TranslateFileName(interp,(char *)ab,&buf);
  if(newname!=NULL) {
    filename=newname;
    Tcl_DStringFree(&buf);
  }
}

// A slightly portable version of fopen
FILE *Nb_FOpen(const char *path,const char *mode)
{
  Nb_DString filename(path);
  Nb_TclNativeFilename(filename);
  return fopen(filename.GetStr(),mode);
}


